#!/usr/bin/python
################################################
#IOPUT v 1.0 Date last modified: 18/2/2007
#distributions.py
################################################

################################################
#This program 'IOPUT' (Index Of Patchiness with Unavailable Traps), was designed to
#calculate the index of patchiness described by Brown et al. (2004) when some
#trap stations are unavailable.
#
#The algorithm spreads out captures evenly over the total available number
#of bins (traps/stations) calculating the weighted sum of the movements made
#
#For three captures over ten bins the captures are broken into segments (tenths)
#and split among bins as follows (numbers represent different captures):
#111 111 111 122 222 222 223 333 333 333
#This will always be the most efficient way of distributing the captures even if the
#captures are clustered at one end (eg: 1,1,1,0,0,0,0,0,0,0)
################################################


import sys 
import csv
import permutations
import random

### generateWeights
# returns the index of patchiness value as described above
# requires  filled_bins     (list) number of bins with objects (ie in a line of 10
#                               with 2 captures, 2)
#           bin_locations   (list) the index value that will call each bin that is
#                               not unavailable from the data list
# returns                   (floating point value) the index of patchiness 

def generateWeights(filled_bins, bin_locations): 
  weight = 0
  bin_count = len(bin_locations)
  if bin_count == 0: 
    return 0 
  segment_count = len(filled_bins) 
  for segment, bin in enumerate(filled_bins): # each segment corresponds to a single filled bin
    low, low_fill = divmod(bin_count * segment, segment_count) 
    high, high_fill = divmod(bin_count * (segment + 1), segment_count) 
    if high_fill >= bin_count: # if this is true, the segment fits into a single bin with room to spare.
      weight += abs(bin - bin_locations[high]) * bin_count 
    else:
      weight += abs(bin - bin_locations[low]) * (segment_count - low_fill) # count the start of the segment in its bin
      if high < bin_count: 
        weight += abs(bin - bin_locations[high]) * high_fill 
      for k in range(low + 1, high): 
        weight += abs(bin - bin_locations[k]) * segment_count
  return float(weight)/bin_count 

### findLiveBins
# requires  bin_list        (list) the data
# returns   a tuple containing:
#           live_bins       (list) contains the contents of each available bin
#           bin_locations   (list) the index value that will call each bin that is
#                               not unavailable from the data list

def findLiveBins(bin_list):
  live_bins = []
  bin_locations = []
  for index, item in enumerate(bin_list):
    if item is not None: # if it's a live bin...
      live_bins.append(item) # add the item in it to the list of live bins...
      bin_locations.append(index) # and add its location to the locations list
  return live_bins, bin_locations

### findFilledBins
# requires  live_bins       (list) contains the contents of each available bin
#           bin_locations   (list) the index value that will call each bin that is
#                               not unavailable from the data list (defaults to None
#                               if not specified)
# returns   filled_bins     (list) returns a paired list of the contents of each
#                               available bin and the index value that will call it

def findFilledBins(live_bins, bin_locations = None):
  filled_bins = []
  if bin_locations is None:
    bin_locations = range(len(live_bins))
  for item, location in zip(live_bins, bin_locations):
    filled_bins.extend([location] * item) # add as many filled bins to the list as there are catches in each bin.
  return filled_bins

### getScatterCots
# requires  bin_list        (list) the data
# returns   scatter_cost    (floating point value) the index of patchiness (via generateWeights)

def getScatterCost(bin_list):
  live_bins, bin_locations = findLiveBins(bin_list)
  filled_bins = findFilledBins(live_bins, bin_locations)
  scatter_cost = generateWeights(filled_bins, bin_locations)
  return scatter_cost

### skipFirstIterations
# requires  iterator    (integer) starting point in the data
#           count       (integer) count value (defaults to 1)

def skipFirstIterations(iterator, count = 1):
  for item in iterator:
    if count > 0:
      count -= 1 # count down until I don't have to skip any more
    else:
      yield item # then produce iterator items after that.

### convertToNumbers
# requires  line    (list) a line from the data file

def convertToNumbers(line):
  return [safeRun(int, None, k) for k in line]

### safeRun
# converts '-'s from the data file to None
# requires  _func           (int) line number for the data being parsed in
#           _failure        catches failures (defaults to None)
#           *args           (list) all unnammed optional arguments in the list
#           **named_args    (list) all nammed optional arguments in the list
# returns the data or None

def safeRun(_func, _failure = None, *args, **named_args):
  try:
    return _func(*args, **named_args) # this might fail...
  except:
    return _failure # but if it does, just give me the argument I passed in; I don't really care how it failed.

### runDataFile
# requires  filename    (str) the name of the file
# prints the output (data + results)

def runDataFile(filename):
  families = {}
  for line in skipFirstIterations(csv.reader(file(filename))):
    name, family, data = line[0], line[1], convertToNumbers(line[2:])
    analysis_output = getPatchinessStuff(data)
    print ",".join([str(k) for k in [name, family] + data + list(analysis_output)])

### getPatchinessStuff
# Computes all required values for patchiness calculations
# requires  data    (list) list of data from the CSV
# returns           (list) the results of the analysis

def getPatchinessStuff(data):
  live_bins, bin_locations = findLiveBins(data)
  filled_bins = findFilledBins(live_bins, bin_locations)
  scatter_cost = generateWeights(filled_bins, bin_locations)
  min_scatter = scatter_cost
  max_scatter = scatter_cost
  scatter_sum = 0
  scatters_above = 0
  scatters_below = 0
  scatter_count = 0
  if permutations.countPermutations(live_bins) < 1000000: # this is the maximum number of permutations permitted before the data are randomised
    is_random = False
    for permuted_live_bins in permutations.permutations(live_bins):
      permuted_filled_bins = findFilledBins(permuted_live_bins, bin_locations)
      permuted_scatter_cost = generateWeights(permuted_filled_bins, bin_locations)
      if permuted_scatter_cost < min_scatter:
        min_scatter = permuted_scatter_cost
      if permuted_scatter_cost > max_scatter:
        max_scatter = permuted_scatter_cost
      scatter_sum += permuted_scatter_cost
      scatter_count += 1
      if permuted_scatter_cost < scatter_cost:
        scatters_below += 1
      else:
        scatters_above  += 1
  else:
    is_random = True
    permuted_live_bins = live_bins[:]
    for k in range(50000): # if the data are randomised, only 50000 permutations are calculated as this is much more time consuming
      random.shuffle(permuted_live_bins)
      permuted_filled_bins = findFilledBins(permuted_live_bins, bin_locations)
      permuted_scatter_cost = generateWeights(permuted_filled_bins, bin_locations)
      if min_scatter is None or permuted_scatter_cost < min_scatter:
        min_scatter = permuted_scatter_cost
      if max_scatter is None or permuted_scatter_cost > max_scatter:
        max_scatter = permuted_scatter_cost
      scatter_sum += permuted_scatter_cost
      scatter_count += 1
      if permuted_scatter_cost < scatter_cost:
        scatters_below += 1
      else:
        scatters_above  += 1
  average_scatter = scatter_sum / scatter_count
  if average_scatter == 0:
    patchiness_index = 0
  else:
    patchiness_index = scatter_cost / average_scatter
  return scatter_cost, is_random, average_scatter, max_scatter, min_scatter, patchiness_index, scatters_above, scatters_below

### Self test code
# if called in IDLE outputs tests below
# if program is run from comand line, this is not used

if __name__ == "__main__":
  if len(sys.argv) == 1:
    print "Self testing.  If anything prints False, something has broken."
    print findFilledBins([0,1,1,0,0,0,1,0,0,0]) == [1, 2, 6]
    print findFilledBins([2,2,3,1,1,2,0,2,1,1,0,2,1,0,1,0,1,2,1,1]) == [0, 0, 1, 1, 2, 2, 2, 3, 4, 5, 5, 7, 7, 8, 9, 11, 11, 12, 14, 16, 17, 17, 18, 19]
    
    print findLiveBins([1,0,0, None,2]) == ([1,0,0,2],[0,1,2,4])
    
    print getScatterCost([0,1,1,0,0,0,1,0,0,0]) == 5.1
    print getScatterCost([0,1,0,0,1,1,0,0,1,0]) == 2.8
    print getScatterCost([1,1,0,0,0,0,0,0,0,0]) == 8.0
    print getScatterCost([1,0,0,0,0,0,0,0,0,1]) == 4.0
    print getScatterCost([0,1,1,0,1,1,1,1,0,0,0,1,1,0,1,1,0,1,1,1,1,1,1,1,0,1,1,1,1,1,1,0,1,1,0,0,0,1,1,1]) == 29.8
    print getScatterCost([2,2,3,1,1,2,0,2,1,1,0,2,1,0,1,0,1,2,1,1]) == 38.2
    print getScatterCost([3,1,3,0,3]) == 3.0
    print getScatterCost([2,2,1,1]) == 2.0
    print getScatterCost([None, None, None, None, None]) == 0.0
    print getScatterCost([1,0,0, None,2]) == 3.25
    
    print convertToNumbers(['0','1','2','3','4','-'])
    
    runDataFile("shortsample.csv")
  else:
    runDataFile(sys.argv[1])
