#!/usr/bin/python

################################################
#IOPUT v 1.0 Date last modified: 18/2/2007
#permutations.py
################################################

################################################
#This program 'IOPUT' (Index Of Patchiness with Unavailable Traps), was designed to
#calculate the index of patchiness described by Brown et al. (2004) when some
#trap stations are unavailable.
#
#This section of the algorithm calculates the permutations, essentially every possible
#ordering of a list of numbers.
################################################

import time

### permutations
# requires  d   (list) the list of data

def permutations(d): 
  data = d[:] 
  data.sort() 
  n = len(data) - 1 
  while 1: 
    yield data 
    j = n - 1 
    while j >= 0 and data[j] >= data[j + 1]: # find the first data point that isn't in descending order from the end
      j -= 1 
    if j < 0: 
      break 
    l = n 
    while data[j] >= data[l]: # find the smallest point in the descending section that is bigger than point found earlier 
      l -= 1 
    swap(data, l, j)
    reverse_slice(data, j + 1) 

### swap
# requires  data    (list) the list of data
#           first   (int) the index of the first point to be swapped
#           second  (int) the index of the second point to be swapped

def swap(data, first, second): # swap two things in a list
  temp = data[first] # create a copy of the first item 
  data[first] = data[second] # make the first item equal to the second
  data[second] = temp #make the second item equal to the copy of the first one (items are now swapped)

### reverse_slice
# requires  data    (list)          the list of data to have a piece reversed
#           start   (int)           the start of the section to be reversed (defaults to 0)
#           end     (int or None)   the end of the section to be reversed (defaults to None)

def reverse_slice(data, start = 0, end = None): 
  if start < 0: 
    first = start + len(data) 
  else: 
    first = start
  if end is None: 
    second = len(data) - 1 
  elif end < 0:
    second = end + len(data) - 1 
  else:
    second = end - 1 
  while first < second: 
    swap(data, first, second) 
    first += 1
    second -= 1 

### countPermutations
# requires  d           (list) the list of data
# returns   numerator   (int) the number of permutations required

def countPermutations(d): 
  item_counts = {}
  for item in d:
    item_counts[item] = item_counts.get(item, 0) + 1
  n = len(d) 
  numerator = 1 
  denominator = 1
  for value in item_counts.itervalues():
    for k in range(value, 0, -1):
      numerator *= n
      denominator *= k
      gcd = greatestCommonDenominator(numerator, denominator)
      numerator /= gcd
      denominator /= gcd
      n -= 1
  return numerator 

### greatestCommonDenominator
# requires  a   (int) first number greatest common denominator required for
#           b   (int) second number greatest common denominator required for
# returns   a   (int) the greatest common denominator 

def greatestCommonDenominator(a, b): 
  while b:
    n = a % b 
    a = b 
    b = n 
  return a

### Time to generate items
# if called in IDLE outputs the time from below
# if program is run from comand line, this is not used

if __name__ == "__main__": # this section estimates the time needed to run the module
  data = [0,0,1,1,0]
  t = time.time()
  x = 0
  for p in permutations([0,0,1,1,0]):
    #print p
    x += 1
  total_time = time.time() - t
  print '%i items in %0.6f seconds: %0.6f seconds per item' % (x, total_time, total_time/x)
  #print data
